/***********************************************************************************************
ReplayToChallenge - Extracts the Challenge from a given Replay file
Copyright (c) 2010, 2011 by Electron

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

This program includes code from the LZO real-time data compression
library. This portion of code is also subject to its own copyright
and licensing conditions, which can be found in the miniLZO directory.

Revision History:
2010-09-12 : Electron : Original version.
2010-09-15 : Electron : Use of shlwapi library.
2010-09-23 : Electron : Advanced error handling.
2011-03-08 : Electron : Minor bug fix.

***********************************************************************************************/

#ifndef STRICT
#define STRICT 1
#endif
#define WIN32_LEAN_AND_MEAN
#define _CRT_SECURE_NO_DEPRECATE

#include <windows.h>
#include <windowsx.h>
#include <shellapi.h>
#include <shlwapi.h>
#include <tchar.h>

#include "minilzo\minilzo.h"

/**********************************************************************************************/

/* Magic 'GBX' */
#define GBX_MAGIC 0x584247

/* Class ID's */
#define GBX_CHALLENGE_TMF 0x03043000
#define GBX_AUTOSAVE_TMF  0x03093000
#define GBX_CHALLENGE_TM  0x24003000
#define GBX_AUTOSAVE_TM   0x2403F000
#define GBX_REPLAY_TM     0x2407E000
#define GBX_CHALLENGE_VSK 0x21080000
#define GBX_REPLAY_VSK    0x2108D000
#define GBX_REGATTA_VSK   0x21002000

/**********************************************************************************************/

const TCHAR g_szGamebox[]   = TEXT(".Gbx");
const TCHAR g_szReplay[]    = TEXT(".Replay.Gbx");
const TCHAR g_szChallenge[] = TEXT(".Challenge.Gbx");
const TCHAR g_szTitle[]     = TEXT("Replay  Challenge");
const TCHAR g_szUsage[]     = TEXT("Usage: ReplayToChallenge.exe <Replay.Gbx>");
const TCHAR g_szErrOom[]    = TEXT("Out of memory.");
const TCHAR g_szErrSrc[]    = TEXT("Source file not found.");
const TCHAR g_szErrOpen[]   = TEXT("Could not open source file.");
const TCHAR g_szErrCreate[] = TEXT("Could not create destination file.");
const TCHAR g_szErrRead[]   = TEXT("Error while reading source file.");
const TCHAR g_szErrWrite[]  = TEXT("Error while writing destination file.");
const TCHAR g_szErrFormat[] = TEXT("Invalid or unsupported file format.");
const TCHAR g_szErrDecomp[] = TEXT("Could not decompress data section.");

#define MsgBox(Text) MessageBox(NULL, Text, g_szTitle, MB_OK | MB_ICONEXCLAMATION)

/**********************************************************************************************/

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	TCHAR szInputPath[MAX_PATH];
	TCHAR szOutputPath[MAX_PATH];
	HANDLE hInputFile = NULL;
	HANDLE hOutputFile = NULL;
	PBYTE pInputData = NULL;
	PBYTE pOutputData = NULL;
	DWORD dwData = 0;
	DWORD dwFileSize = 0;
	DWORD dwHeaderSize = 0;
	DWORD dwNumOfBytesRead = 0;
	DWORD dwNumOfBytesWritten = 0;
	DWORD dwUncompressedSize = 0;
	DWORD dwCompressedSize = 0;
	PTSTR pch = NULL;

	/* The source file must be passed as program argument */
	if (lpCmdLine == NULL || lpCmdLine[0] == '\0')
	{
		MsgBox(g_szUsage);
		return 0;
	}

	/* Define the symbols UNICODE and _UNICODE to support Unicode filenames */
#ifdef UNICODE
	{
		int argc;
		LPTSTR* argv = CommandLineToArgvW(GetCommandLine(), &argc);
		if (argv == NULL || argc <= 1)
			return 0;

		_tcsncpy(szInputPath, argv[1], MAX_PATH-1);
	}
#else
	_tcsncpy(szInputPath, (LPCTSTR)lpCmdLine, MAX_PATH-1);
#endif

	/* Remove quotation marks */
	PathUnquoteSpaces(szInputPath);

	/* Does the passed file exist? */
	if (!PathFileExists(szInputPath))
	{
		MsgBox(g_szErrSrc);
		return 0;
	}

	/* Build the name of the destination file */
	_tcsncpy(szOutputPath, szInputPath, MAX_PATH-1);
	if (pch = StrStrI(szOutputPath, g_szReplay))
		*pch = TEXT('\0');
	else if (pch = StrStrI(szOutputPath, g_szGamebox))
		*pch = TEXT('\0');
	_tcsncat(szOutputPath, g_szChallenge, MAX_PATH);

	/* Open source file */
	hInputFile = CreateFile(szInputPath, GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hInputFile == INVALID_HANDLE_VALUE)
	{
		MsgBox(g_szErrOpen);
		return 0;
	}

	/* Retrieve file size */
	dwFileSize = GetFileSize(hInputFile, NULL);

	/* Read and check magic 'GBX' */
	if (!ReadFile(hInputFile, (LPVOID)&dwData, 3, &dwNumOfBytesRead, NULL) ||
		dwNumOfBytesRead != 3 || (dwData & 0x00FFFFFF) != GBX_MAGIC)
	{
		MsgBox(g_szErrFormat);
		goto Exit;
	}

	/* Jump to storage settings */
	if (SetFilePointer(hInputFile, 2, NULL, FILE_CURRENT) == 0xFFFFFFFF)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Read storage settings */
	if (!ReadFile(hInputFile, (LPVOID)&dwData, 4, &dwNumOfBytesRead, NULL) ||
		dwNumOfBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Is this an uncompressed gamebox file? */
	if ((dwData & 0x00FF0000) == 0x00550000)
	{
		MsgBox(g_szErrFormat);
		goto Exit;
	}

	/* Read Class Id */
	if (!ReadFile(hInputFile, (LPVOID)&dwData, 4, &dwNumOfBytesRead, NULL) ||
		dwNumOfBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Is this a replay? */
	if (dwData != GBX_REPLAY_TM && dwData != GBX_AUTOSAVE_TMF &&
		dwData != GBX_AUTOSAVE_TM && dwData != GBX_REPLAY_VSK)
	{
		MsgBox(g_szErrFormat);
		goto Exit;
	}

	/* Read header size */
	if (!ReadFile(hInputFile, (LPVOID)&dwHeaderSize, 4, &dwNumOfBytesRead, NULL) ||
		dwNumOfBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Jump directly to the block size information in the data section */
	if (SetFilePointer(hInputFile, dwHeaderSize + 8, NULL, FILE_CURRENT) == 0xFFFFFFFF)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Read size of uncompressed data section */
	if (!ReadFile(hInputFile, (LPVOID)&dwUncompressedSize, 4, &dwNumOfBytesRead, NULL) ||
		dwNumOfBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Read size of compressed data section */
	if (!ReadFile(hInputFile, (LPVOID)&dwCompressedSize, 4, &dwNumOfBytesRead, NULL) ||
		dwNumOfBytesRead != 4)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Sanity check */
	if (dwCompressedSize > dwFileSize)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Allocate memory for the compressed data */
	pInputData = GlobalAllocPtr(GHND, dwCompressedSize);
	if (pInputData == NULL)
	{
		MsgBox(g_szErrOom);
		goto Exit;
	}

	/* Allocate memory for the uncompressed data */
	pOutputData = GlobalAllocPtr(GHND, dwUncompressedSize);
	if (pOutputData == NULL)
	{
		MsgBox(g_szErrOom);
		goto Exit;
	}

	/* Read entire data section */
	if (!ReadFile(hInputFile, pInputData, dwCompressedSize, &dwNumOfBytesRead, NULL) ||
		dwNumOfBytesRead != dwCompressedSize)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Initialize the LZO library */
	if (lzo_init() != LZO_E_OK)
	{
		MsgBox(g_szErrDecomp);
		goto Exit;
	}

	/* Decompress the data section */
	dwNumOfBytesRead = dwUncompressedSize;
	if (lzo1x_decompress_safe(pInputData, dwCompressedSize, pOutputData,
		&dwNumOfBytesRead, NULL) != LZO_E_OK || dwNumOfBytesRead != dwUncompressedSize)
	{
		MsgBox(g_szErrDecomp);
		goto Exit;
	}

	/* Read the Id of the first chunk */
	CopyMemory(&dwData, pOutputData, sizeof(dwData));

	/* Check the Chunk Id */
	if ((dwData & 0x00000FFF) != 0x002)
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Read the size of the embedded challenge */
	CopyMemory(&dwFileSize, pOutputData + 4, sizeof(dwFileSize));

	/* Sanity check */
	if (dwFileSize > (dwNumOfBytesRead - 8))
	{
		MsgBox(g_szErrRead);
		goto Exit;
	}

	/* Create destination file */
	hOutputFile = CreateFile(szOutputPath, GENERIC_READ | GENERIC_WRITE,
		0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hOutputFile == INVALID_HANDLE_VALUE)
	{
		MsgBox(g_szErrCreate);
		goto Exit;
	}

	/* Write challenge data */
	if (!WriteFile(hOutputFile, pOutputData + 8, dwFileSize, &dwNumOfBytesWritten, NULL) ||
		dwNumOfBytesWritten != dwFileSize)
	{
		CloseHandle(hOutputFile);
		hOutputFile = NULL;
		DeleteFile(szOutputPath);
		MsgBox(g_szErrWrite);
		/* Fall through */
	}

Exit:
	/* Cleanup */
	if (pOutputData != NULL)
		GlobalFreePtr(pOutputData);
	if (pInputData != NULL)
		GlobalFreePtr(pInputData);
	if (hOutputFile != NULL)
		CloseHandle(hOutputFile);
	if (hInputFile != NULL)
		CloseHandle(hInputFile);

	return 0;
}

/**********************************************************************************************/
